// Persistence of Vision Ray Tracer Scene Description File
// File: bsplines.inc
// Vers: 3.5
// Desc: Macro Collection for Bezier-Splines
// Date: 28.03.02
// Auth: Tim Nikias Wenclawiak

/*Macros:
Spline-Modifiers (no return-value):
BSpline_Translate  => give BSpline and translation vector => modifies BSpline
BSpline_Flip       => reverses the given BSpline => modifies BSpline
BSpline_Scale      => give BSpline and scale vector => modifies BSpline
BSpline_Rotate     => give BSplne and rotation vector => modifies BSpline

CSG-Generators (return bunch of objects):
Show_BSpline       => give BSpline, radius and CSG-Amount => returns cylinders and spheres
Morph_Show_BSpline => give 2 BSplines, clock, radius and CSG_Amount => returns cylinders and spheres
Show_BNodes        => give BSpline and radius => returns spheres (placed on nodes)
Show_BDirs         => give BSpline and radius => returns cones (indicating direction of bspline)

Calculators (return vectors):
BSpline_Pos        => give BSpline and mover => returns position
BSpline_Normal     => give BSpline and mover => returns curvature-normal

Extra:
Linear_Pos         => give Linear-Interpolation and mover => returns position
BSpline2Linear      => give BSpline, steps, final node-amount, filename => writes file

With "mover", a float between 0 and 1 (beginning to end of spline) is meant.
Linear_Pos is just an algorithm to connect an array of vectors. With BSpline2Linear
you can generate and write such an array. Useful to just have a series of points
to be interpolated smoothly (e.g. no change in speed due to different curvatures etc).

The file generated by BSpline2Linear has two arrays:
Bezier_Node : The original BSpline
Linear_Node : A series of points that lie on the BSpline
*/


#declare _bsplines_inc_tnw=1; //For other include to check if macros were included

//Just a detail value, no actual need to change it
#declare Zero_Value=.0001;

/***********************/
/*****BSpline_Translate**/
/***********************/
#macro BSpline_Translate(_bspline_array,Translation)
 #local Saved_Length=dimension_size(_bspline_array,1);
  //Loop that jumps from position-node to position-node (C+2)
  #local C=0; #while (C<Saved_Length)
   #declare _bspline_array[C]=_bspline_array[C]+Translation;  
  #local C=C+2; #end
#end

/***********************/
/*******BSpline_Flip*****/
/***********************/
#macro BSpline_Flip(_bspline_array)
 #local Saved_Length=dimension_size(_bspline_array,1);
 #local Old_Spline=_bspline_array 
  #local C=0; #while (C<Saved_Length)
   #declare _bspline_array[C]=Old_Spline[Saved_Length-2-C];
   #declare _bspline_array[C+1]=-Old_Spline[Saved_Length-1-C];
  #local C=C+2; #end
#end

/***********************/
/*****BSpline_Scale******/
/***********************/
#macro BSpline_Scale(_bspline_array,Scalement)
 #local Saved_Length=dimension_size(_bspline_array,1);
  //Loop that jumps from position-node to position-node (C+2)
  #local C=0; #while (C<Saved_Length)
   #declare _bspline_array[C]=_bspline_array[C]*Scalement;  
  #local C=C+1; #end
#end

/***********************/
/*****BSpline_Rotate*****/
/***********************/
#macro BSpline_Rotate(_bspline_array,Rotation)
 #local Saved_Length=dimension_size(_bspline_array,1);
  //Loop that jumps from position-node to position-node (C+2)
  #local C=0; #while (C<Saved_Length)
   #declare _bspline_array[C]=vrotate(_bspline_array[C],Rotation);  
  #local C=C+1; #end
#end


/***********************/
/*******Show_BSpline*****/
/***********************/
#macro Show_BSpline(__bspline_array,_Thickness,_Detail)
  //Loop running through the different nodes
  #local _Slice_Val = 0;
  #while (_Slice_Val <= _Detail)
   
   #local _Actual_Pos = BSpline_Pos(__bspline_array,(_Slice_Val/_Detail));
    
   sphere{_Actual_Pos,_Thickness}
    
   #ifdef (_Former_Pos)
    #if (vlength(_Former_Pos-_Actual_Pos)!=0)
     cylinder{_Former_Pos,_Actual_Pos,_Thickness}
    #end
   #end
   
  #local _Former_Pos = _Actual_Pos;
  //End of loop
  #local _Slice_Val = _Slice_Val+1;
 #end
#end

/*************************/
/****Morph_Show_BSpline****/
/*************************/
#macro Morph_Show_BSpline(_bspline_array_1,_bspline_array_2,Clock,Thickness,Detail)
  //Loop running through the different nodes
  #local Slice_Val = 0;
  #while (Slice_Val <= Detail)
   #local Actual_Pos_1 = BSpline_Pos(_bspline_array_1,Slice_Val/Detail);
   #local Actual_Pos_2 = BSpline_Pos(_bspline_array_2,Slice_Val/Detail);
    #local Actual_Pos=Actual_Pos_1-(Actual_Pos_1-Actual_Pos_2)*Clock;
    sphere{Actual_Pos,Thickness}
    #ifdef (Former_Pos)
     #if (vlength(Former_Pos-Actual_Pos)!=0)
     cylinder{Former_Pos,Actual_Pos,Thickness}
     #end
    #end
   #local Former_Pos = Actual_Pos;
  //End of loop
  #local Slice_Val = Slice_Val+1;
 #end
#end

/***********************/
/*******Show_BNodes******/
/***********************/
#macro Show_BNodes(_bspline_array,Thickness)
 #local Nodes = dimension_size(_bspline_array,1);
 #local Node_Val=0;
 #while (Node_Val < Nodes)
  sphere{_bspline_array[Node_Val],Thickness}
 #local Node_Val=Node_Val+2;
 #end
#end

/***********************/
/*******Show_BDirs******/
/***********************/
#macro Show_BDirs(_bspline_array,Thickness)
 #local Nodes = dimension_size(_bspline_array,1);
 #local Node_Val=0;
 #while (Node_Val < Nodes)
  cone{_bspline_array[Node_Val],Thickness _bspline_array[Node_Val]+_bspline_array[Node_Val+1],0}
 #local Node_Val=Node_Val+2;
 #end
#end

/***********************/
/******BSpline_Pos*******/
/***********************/
//Calculates my Bezier-Spline-Algorithm
#macro BSpline_Pos(_b_spl_array,_nm)
 //Parameter: Bezier-Spline-Array, Node_Mover (0->1)
 //Amount of interpolations
 #local _nodes = dimension_size(_b_spl_array,1)/2;
 #local _max_nm=off;
 #local _a_nm=_nm;
 //Node_Mover-Corrections, keeps them in boundaries
 #if (_a_nm >= 1) #local _a_nm=1; #local _max_nm=on; #end
 #if (_a_nm < 0) #local _a_nm=0; #end
 
 //Index of section:
 #local _i_sec=floor(_a_nm*(_nodes-1))*2;
 //Actual section's interpolation-value (nodemover)
 #local _a_nm=_a_nm*(_nodes-1)-floor(_a_nm*(_nodes-1));

 #if (_max_nm)
  #local _Bezier_Pos=_b_spl_array[_nodes*2-2]+_b_spl_array[_nodes*2-1]*.01;
 #else
  #local _Bezier_Pos=
   _b_spl_array[_i_sec]*pow(1-_a_nm,3)+
   3*(_b_spl_array[_i_sec+1]+_b_spl_array[_i_sec])*pow(1-_a_nm,2)*_a_nm-
   3*(_b_spl_array[_i_sec+3]-_b_spl_array[_i_sec+2])*(1-_a_nm)*pow(_a_nm,2)+
   _b_spl_array[_i_sec+2]*pow( _a_nm ,3);
 #end
 //Return the Position-Vector
 _Bezier_Pos
#end

/***********************/
/******Linear_Pos*******/
/***********************/
//Usage: Call Macro with Linear-I-Array and Mover between 0 and 1 (beginning and end of spline)
#macro Linear_Pos(_Linear_Array,_LI_Mover)
 //Amount of Interpolations
 #local _Linear_Sections = dimension_size(_Linear_Array,1)-1;
 //Node_Mover-Corrections, keeps them in boundaries
 #if (_LI_Mover > 1) #local _LI_Mover=1; #end
 #if (_LI_Mover < 0) #local _LI_Mover=0; #end

 #if (_LI_Mover = 0)
  #local _Linear_Pos_ = _Linear_Array[0];
 #end
 #if (_LI_Mover = 1)
  #local _Linear_Pos_ = _Linear_Array[_Linear_Sections];
 #end
 #if (_LI_Mover > 0 & _LI_Mover < 1)
  #local _Actual = floor(_LI_Mover*_Linear_Sections);
  #local _Next = floor(_LI_Mover*_Linear_Sections)+1;
  #local _Time = _LI_Mover*(_Linear_Sections)-_Actual;
  #local _Linear_Pos_ = _Linear_Array[_Actual]+(_Linear_Array[_Next]-_Linear_Array[_Actual])*_Time;
 #end
_Linear_Pos_
#end

/***********************/
/*****Spline2Linear*****/
/***********************/
#macro BSpline2Linear(_b_bspline_array,Calc_Detail,Final_Nodes,Filename)

//The created Include-File with saved Spline-Data
#if (strcmp(Filename,"splines.inc") = 0)
 #debug "\n*!* Warning, splines.inc reserved! *!*\n*!* Renamed to spline.spl!\n"
 #declare Filename = "spline.spl"
#end

//Fragmenting the original Spline and calculating the length...
#declare Sm_Array = array[Calc_Detail+1]
#declare Bezier_Length = 0;
#declare Sm_Array[0]=BSpline_Pos(_b_bspline_array,0);

#declare Smooth_Check = 1;
#while (Smooth_Check <= Calc_Detail)
 #declare Sm_Array[Smooth_Check]=BSpline_Pos(_b_bspline_array,Smooth_Check/Calc_Detail);
 #declare Bezier_Length = Bezier_Length+(vlength(Sm_Array[Smooth_Check]-Sm_Array[Smooth_Check-1]));
#declare Smooth_Check = Smooth_Check+1;
#end

//We have the Length and we have an array of vectors. What do we do now?
//Using the Length and the Final_Nodes-Value, a length is calculated
//which defines the distance covered by every section. At those lengths, the
//position is saved. Voila! We have our new pseudo-Spline.

#declare Section_Length = Bezier_Length/Final_Nodes;
#declare PS_Spline_Node = array[Final_Nodes+1]

//First point still needs to be first point, as well as last one 
 #declare PS_Spline_Node[0]=Sm_Array[0];
 #declare PS_Spline_Node[Final_Nodes]=Sm_Array[Calc_Detail];

//Length_Check is a integer which keeps track of the fragments passed, in
//order to never use two same nodes. The Length keeps track of the distance
//spanned between two fragments, supplying a method of tracking new pseudo-nodes.

#declare Length = 0;
#declare Length_Check = 1;
#declare Node_Val = 1; 
#while (Length_Check <= Calc_Detail)
 #declare Length = Length+(vlength(Sm_Array[Length_Check]-Sm_Array[Length_Check-1]));
 //If-Section which covers the point at which new Node is found
 #if (Length >= Section_Length)
  #declare PS_Spline_Node[Node_Val]=Sm_Array[Length_Check];
  #declare Node_Val = Node_Val+1;
  #declare Length = Length-Section_Length;
 #end
 //Preparing the loop to move on until all Fragments are passed
 #declare Length_Check = Length_Check+1;
#end

/* Output */
#fopen Spline Filename write
#write (Spline "//File created by Tim Wenclawiak's Spline2Linear-Macro\n")
#write (Spline "//Original Bezier-Spline\n")
//The original-Spline
#write (Spline "#declare Bezier_Node = array[",str(dimension_size(_b_bspline_array,1),1,0),"]\n{\n")
#declare Write = 0;
#while (Write < dimension_size(_b_bspline_array,1))
#write (Spline _b_bspline_array[Write],"\n")
#declare Write=Write+1;
#end
#write (Spline "}\n")

//The pseudo-Spline
#write (Spline "\n//Linear-Interpolation-Data\n")

#write (Spline "#declare Linear_Node = array[",str(Final_Nodes+1,1,0),"]\n{\n")
#declare Write = 0;
#while (Write <= Final_Nodes)
#write (Spline PS_Spline_Node[Write],"\n")
#declare Write=Write+1;
#end
#write (Spline "}\n")

//The Length of the Spline
#write (Spline "\n//Bezier-Length\n")
#write (Spline "#declare Bezier_Length = ",Bezier_Length,";\n")

/* End of Output */
#fclose Spline

/*End of Macro*/

#end


/***********************/
/*****BSpline_Normal****/
/***********************/
//BSpline_Normal
//Usage: Call Macro with Spline and Node-Mover, returns normal-vector
#macro BSpline_Normal(_spl_array,_nm)

 #local BF_Node_Mover = _nm;
 //Calculating some needed positions on the spline
  #if (BF_Node_Mover >= Zero_Value)
   #local Node_M_1 = BF_Node_Mover-Zero_Value;
   #local Node_M_2 = BF_Node_Mover;
   #local Node_M_3 = BF_Node_Mover+Zero_Value;
  #end
  #if (BF_Node_Mover < Zero_Value)
   #local Node_M_1 = 0;
   #local Node_M_2 = BF_Node_Mover+pow(Zero_Value,2);
   #local Node_M_3 = BF_Node_Mover+Zero_Value;
   #end
  #if (BF_Node_Mover >= 1-Zero_Value)
   #local Node_M_1 = BF_Node_Mover-Zero_Value;
   #local Node_M_2 = BF_Node_Mover-pow(Zero_Value,2);
   #local Node_M_3 = 1;
  #end
 //Use calculated Node_Movers and calculate Positions
  #local Centri_Pos_1 = BSpline_Pos(_spl_array,Node_M_1);
  #local Centri_Pos_2 = BSpline_Pos(_spl_array,Node_M_2);
  #local Centri_Pos_3 = BSpline_Pos(_spl_array,Node_M_3);
   #local _veca=vnormalize(Centri_Pos_2-Centri_Pos_1);
   #local _vecb=vnormalize(Centri_Pos_3-Centri_Pos_2);
   
   //Begin with assumption that we don't need tp do extra calculations
   #local Again=false;
   //Same or inverse? If so, we do some extra calculations
   #if ((_veca.x = _vecb.x & _veca.y = _vecb.y & _veca.z = _vecb.z) | (_veca.x = -_vecb.x & _veca.y = -_vecb.y & _veca.z = -_vecb.z))
    #local Again=true;
   #end
  
  //In Loop, do steps in relation to actual Bezier-Spline
  #local Detail=1/(dimension_size(_spl_array,1)/2)*.05;
  //Counter to limit the loop
  #local C=0;
  #while (Again) 
   //Dependant on where we are, we move different node-movers and re-calculate appropriate positions/vectors
   #if (BF_Node_Mover >= Zero_Value)
    #local Node_M_3 = Node_M_3+Detail;
    #local Centri_Pos_3 = BSpline_Pos(_spl_array,Node_M_3);
    #local _vecb=vnormalize(Centri_Pos_3-Centri_Pos_2);
   #end
   #if (BF_Node_Mover < Zero_Value)
    #local Node_M_3 = Node_M_3+Detail;
    #local Centri_Pos_3 = BSpline_Pos(_spl_array,Node_M_3);
    #local _vecb=vnormalize(Centri_Pos_3-Centri_Pos_2);
   #end
   #if (BF_Node_Mover >= 1-Zero_Value)
    #local Node_M_1 = Node_M_1-Detail;
    #local Centri_Pos_1 = BSpline_Pos(_spl_array,Node_M_1);
    #local _veca=vnormalize(Centri_Pos_2-Centri_Pos_1);
   #end
      
   //Same or inverse again?
   #if ((_veca.x = _vecb.x & _veca.y = _vecb.y & _veca.z = _vecb.z) | (_veca.x = -_vecb.x & _veca.y = -_vecb.y & _veca.z = -_vecb.z))
    #local Again=true; #else
    #local Again=false; #end
   //Increment the counter
   #local C=C+1;
   //Finally, check if we have already looped to much (normally, 25 should be max!)
   #if (C=50) #local Again=false; #end
  #end

 #local _Centri_Vec_=vnormalize(vcross(vcross(Centri_Pos_2-Centri_Pos_1,Centri_Pos_2-Centri_Pos_3),Centri_Pos_2-Centri_Pos_1));
 //Return the normal
 (_Centri_Vec_)
#end
